import networkx as nx

from .rxn_graph import build_reaction_graph, define_node_edge_matches

# Define node and edge match only once at import
_NODE_MATCH, _EDGE_MATCH = define_node_edge_matches()


def are_atom_maps_equivalent(
    gt_smi: str, pred_smi: str, canonicalize: bool = False
) -> bool:
    """
    Compare two atom-mapped reaction SMILES to determine whether
    the atom mappings are equivalent.

    This comparison is based on:
    - Canonicalizing the input SMILES (Optional)
    - Constructing reaction graphs
    - Evaluating graph isomorphism with matching atom/bond features

    Args:
        gt_smi (str): Ground truth atom-mapped reaction SMILES
        pred_smi (str): Predicted atom-mapped reaction SMILES
        canonicalize (bool): Whether to canonicalize both before evaluation

    Returns:
        bool: True if the atom mappings are equivalent, False otherwise
    """
    rxn_gt_G = build_reaction_graph(gt_smi, canonicalize=canonicalize)
    rxn_pred_G = build_reaction_graph(pred_smi, canonicalize=canonicalize)

    return nx.is_isomorphic(
        rxn_gt_G, rxn_pred_G, node_match=_NODE_MATCH, edge_match=_EDGE_MATCH
    )



import csv
import json
from typing import List

from .data_models import ReactionPair

"""
# Example CSV
id,ground_truth,prediction
rxn1,[CH3:1][OH:2]>>[CH3:1][O-:2],[Na+:3].[CH3:1][OH:2]>>[CH3:1][O-:2].[Na+:3]
rxn2,[C:1](=[O:2])[O-:3]>>[C:1](=[O:2])[OH:3],[C:1](=[O:2])[OH:3]>>[C:1](=[O:2])[OH:3]

# Example usage
from atommap_eval.input_io import load_pairs_from_csv
from atommap_eval.parallel import evaluate_pairs_in_parallel

pairs = load_pairs_from_csv("reactions.csv")
results = evaluate_pairs_in_parallel(pairs)

"""


def load_pairs_from_csv(
    path: str,
    gt_col: str = "ground_truth",
    pred_col: str = "prediction",
    id_col: str = "id",
) -> List[ReactionPair]:
    """
    Load reaction pairs from a CSV file and return a list of ReactionPair objects.

    Args:
        path (str): Path to CSV file.
        gt_col (str): Name of the column containing ground truth SMILES.
        pred_col (str): Name of the column containing predicted SMILES.
        id_col (str): Optional column for identifying reactions.

    Returns:
        List[ReactionPair]
    """
    pairs = []
    with open(path, newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            gt = row[gt_col].strip()
            pred = row[pred_col].strip()
            rxn_id = row.get(id_col, None)
            pairs.append(ReactionPair(gt, pred, rxn_id))
    return pairs


def load_pairs_from_json(path: str) -> List[ReactionPair]:
    """
    Load reaction pairs from a JSON file containing a list of dicts with keys:
    "ground_truth", "prediction", and optionally "id".

    Args:
        path (str): Path to JSON file.

    Returns:
        List[ReactionPair]
    """
    with open(path, "r") as f:
        data = json.load(f)

    if not isinstance(data, list):
        raise ValueError("Expected a list of dictionaries in the JSON file.")

    pairs = []
    for entry in data:
        gt = entry["ground_truth"]
        pred = entry["prediction"]
        rxn_id = entry.get("id")
        pairs.append(ReactionPair(gt, pred, rxn_id))

    return pairs



from concurrent.futures import ProcessPoolExecutor, TimeoutError, as_completed
from typing import List, Optional, Tuple, Union
from functools import partial
import time
import multiprocessing as mp

from wrapt_timeout_decorator import timeout

from .data_models import ReactionPair
from .evaluator import are_atom_maps_equivalent



def _unpack_pair(pair: Union[Tuple[str, str], "ReactionPair"]) -> Tuple[str, str]:
    """Extract (gt, pred) from a tuple or a ReactionPair."""
    if isinstance(pair, tuple):
        return pair
    elif hasattr(pair, "ground_truth") and hasattr(pair, "prediction"):
        return pair.ground_truth, pair.prediction
    else:
        raise TypeError(f"Unsupported pair type: {type(pair)}")


def _safe_equivalence_check(gt: str, pred: str, canonicalize: bool = False) -> bool:
    """Safely call are_atom_maps_equivalent in a subprocess."""
    return are_atom_maps_equivalent(gt_smi=gt, pred_smi=pred, canonicalize=canonicalize)


@timeout(dec_timeout=10.0, use_signals=False)
def timed_are_maps_equivalent(gt, pred, canonicalize):
    return are_atom_maps_equivalent(gt_smi=gt, pred_smi=pred, canonicalize=canonicalize)


def safe_timed_equivalence_check(gt: str, pred: str, canonicalize: bool=False):
    try:
        result = timed_are_maps_equivalent(gt, pred, canonicalize)
        return result, "ok"
    except TimeoutError:
        return None, "timeout"
    except Exception as e:
        return None, f"error:{type(e).__name__}"
    

def evaluate_pairs_batched(
    pairs: List[Union[Tuple[str, str], "ReactionPair"]],
    canonicalize: bool = False,
    num_workers: Optional[int] = None,
    batch_size: int = 1000,
) -> List[Tuple[Optional[bool], str]]:
    """
    Evaluate atom-map equivalence for many (gt, pred) pairs using batches, timeouts, and parallelism.

    Args:
        pairs: list of (gt_smiles, pred_smiles) tuples or ReactionPair objects
        canonicalize: whether to canonicalize before comparison
        num_workers: number of parallel workers (defaults to os.cpu_count())
        batch_size: number of reactions to process per batch
        timeout: per-reaction timeout in seconds

    Returns:
        List[Tuple[Optional[bool], str]]:
            Each element is (equivalent, status)
            where status is in {"ok", "timeout", "invalid_input", "error:<type>"}
    """
    results_all: List[Tuple[Optional[bool], str]] = []
    total_start = time.perf_counter()

    print(f"Starting evaluation of {len(pairs)} pairs with {num_workers or 'default'} workers...", flush=True)

    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        for start in range(0, len(pairs), batch_size):
            end = min(start + batch_size, len(pairs))
            print(f"Submitting batch {start}-{end} / {len(pairs)}...", flush=True)

            batch = pairs[start:end]
            futures = []

            for pair in batch:
                gt, pred = _unpack_pair(pair)
                if not isinstance(gt, str) or not isinstance(pred, str) or not gt.strip() or not pred.strip():
                    results_all.append((None, "invalid_input"))
                    continue
                futures.append(executor.submit(safe_timed_equivalence_check, gt, pred, canonicalize))

            # collect results
            for future in as_completed(futures):
                try:
                    equivalent, status = future.result()
                except Exception as e:
                    equivalent, status = None, f"error:{type(e).__name__}"
                results_all.append((equivalent, status))

    total_elapsed = time.perf_counter() - total_start
    print(f"[DONE] Processed {len(pairs)} pairs in {total_elapsed:.2f}s using {num_workers or 'default'} workers.")
    return results_all



from dataclasses import dataclass
from typing import Optional, Dict

"""
# Example

from atommap_eval.data_models import ReactionPair

rp = ReactionPair(
    ground_truth="[CH3:1][OH:2]>>[CH3:1][O-:2]",
    prediction="[CH3:1][OH:2]>>[CH3:1][O-:2]",
    id="rxn_42"
)

print(rp.ground_truth)
print(rp.id)

"""


@dataclass(frozen=True)
class ReactionPair:
    """
    Represents a ground-truth / predicted reaction SMILES pair.

    Attributes:
        ground_truth (str): Ground-truth atom-mapped reaction SMILES.
        prediction (str): Predicted atom-mapped reaction SMILES.
        id (Optional[str]): Optional identifier for tracking the reaction
            (e.g., filename, reaction ID).
    """

    ground_truth: str
    prediction: str
    id: Optional[str] = None


@dataclass
class PreprocessResult:
    """
    Represents the preprocessing structure of a ground-truth / predicted reaction SMILES pair.

    Attributes:
        reference: preprocessed ground-truth reaction. None if unfit for evaluation.
        prediction: preprocessed predicted reaction SMILES pair. None if unfit for evaluation.
        flags_ref: notify either why the preprocessing failed or what formatting problems are shown by the ground-truth.
        flags_pred: notify either why the preprocessing failed or what formatting problems are shown by the prediction
    """
    reference: Optional[str]
    prediction: Optional[str]
    flags_ref: Dict[str, bool]
    flags_pred: Dict[str, bool]


import csv
import os
import subprocess
import tempfile


def test_cli_csv_roundtrip():
    with tempfile.TemporaryDirectory() as tmpdir:
        input_path = os.path.join(tmpdir, "input.csv")
        output_path = os.path.join(tmpdir, "output.csv")

        # Write a basic test input
        with open(input_path, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["id", "ground_truth", "prediction"])
            writer.writerow(
                ["rxn1", "[CH3:1][OH:2]>>[CH3:1][O-:2]", "[CH3:1][OH:2]>>[CH3:1][O-:2]"]
            )
            writer.writerow(
                ["rxn2", "[CH3:2][OH:1]>>[CH3:1][O-:2]", "[CH3:9][OH:8]>>[CH3:9][O-:8]"]
            )

        # Run CLI
        result = subprocess.run(
            [
                "python",
                "-m",
                "atommap_eval.cli",
                input_path,
                "-f",
                "csv",
                "-p",
                "2",
                "-o",
                output_path,
            ],
            capture_output=True,
            text=True,
        )

        assert result.returncode == 0
        assert "2 reactions" in result.stdout
        assert os.path.exists(output_path)

        # Check output content
        with open(output_path, newline="") as f:
            rows = list(csv.DictReader(f))
            assert rows[0]["id"] == "rxn1"
            assert rows[0]["equivalent"] == "True"
            assert rows[1]["id"] == "rxn2"
            assert rows[1]["equivalent"] == "False"



import re
import pandas as pd
from rdkit import Chem
from rdkit.Chem import AllChem
from typing import Dict, Iterable, List, Optional, Union

from atommap_eval.data_models import ReactionPair, PreprocessResult
from .rxnmapper_utils import (
    create_rxn_smiles_from_rxn_equation,
    generate_atom_mapped_reaction_atoms,
    process_reaction_with_product_maps_atoms,
)

# Level 0 - preprocessing
def canonicalize_rxn_smiles(rxn_smi: str) -> Optional[str]:
    """
    Canonicalizes a reaction SMILES string, reordering molecules
    and atom-mapping numbers deterministically.

    Args:
        rxn_smi (str): Atom-mapped reaction SMILES string.

    Returns:
        Optional[str]: Canonicalized atom-mapped reaction SMILES,
        or None if it fails.
    """
    try:
        rxn_smi_can_dict = process_reaction_with_product_maps_atoms(rxn_smi)

        if rxn_smi_can_dict is None or not isinstance(rxn_smi_can_dict, tuple):
            raise ValueError(
                f"Invalid output from "
                f"process_reaction_with_product_maps_atoms: {rxn_smi_can_dict}"
            )

        rxn_can, _ = generate_atom_mapped_reaction_atoms(
            *rxn_smi_can_dict, canonical=True
        )
        can_smi = create_rxn_smiles_from_rxn_equation(rxn_can)

        return can_smi
    except Exception as e:
        raise RuntimeError(
            f"Canonicalization failed for input SMILES: {rxn_smi}\nError: {e}"
        )


def sanitize_reaction_smiles(rxn_smi: str) -> Optional[str]:
    """
    Sanitize a reaction SMILES string.
    Returns an isomeric, canonical RDKit-consistent SMILES with proper aromaticity.

    Args:
        rxn_smi (str): Atom-mapped reaction SMILES string.
    
    Returns:
        Optional[str]: Sanitized isomeric, canonical SMILES,
        None if it fails.
    """
    try:
        rxn = AllChem.ReactionFromSmarts(rxn_smi, useSmiles=True)
        if rxn is None:
            raise ValueError(f"Invalid reaction SMILES: {rxn_smi}")
    except Exception:
        return None

    def sanitize_mols(mols):
        sanitized = []
        for mol in mols:
            try:
                if mol is None:
                    continue
                mol = Chem.Mol(mol)
                Chem.SanitizeMol(mol)
                Chem.AssignStereochemistry(mol, cleanIt=True, force=True)
                Chem.SetAromaticity(mol)
                sanitized.append(mol)
            except Exception:
                return None
        return sanitized

    reactants = sanitize_mols(rxn.GetReactants())
    products  = sanitize_mols(rxn.GetProducts())
    agents    = sanitize_mols(rxn.GetAgents())
    if reactants is None or products is None or agents is None:
        return None
    
    def to_smiles(mols):
        return '.'.join(sorted(
            Chem.MolToSmiles(m, canonical=True, isomericSmiles=True) for m in mols
        ))

    react_smiles   = to_smiles(reactants)
    agent_smiles   = to_smiles(agents)
    product_smiles = to_smiles(products)

    return f"{react_smiles}>{agent_smiles}>{product_smiles}"


# Level 0 - formatting diagnostics
def detect_atom_no_mapnum_in_preproc_prod(rxn_smi: str) -> bool:
    """
    Return True if the reaction contains at least one atom in the product side
    that does not have an atom-mapping number (after canonicalization, one atom-map number won't be found in the reactants side).
    Returns False otherwise.
    Returns None if the reaction SMILES cannot be parsed.
    """
    try:
        rxn = AllChem.ReactionFromSmarts(rxn_smi, useSmiles=True)
        if rxn is None:
            return None
    except Exception:
        return None
    set_prod_map = set()
    set_reac_map = set()
    for product in rxn.GetProducts():
        for atom in product.GetAtoms():
            amap = atom.GetAtomMapNum()
            if amap:
                set_prod_map.add(amap)
    for reactant in rxn.GetReactants():
        for atom in reactant.GetAtoms():
            amap = atom.GetAtomMapNum()
            if amap:
                set_reac_map.add(amap)
    
    missing = not set_prod_map.issubset(set_reac_map)
    return missing


# Level 1 - preprocessing
def preprocess_reaction_pair(ref_rxn: str, pred_rxn: str) -> PreprocessResult:
    """
    Run canonicalization + sanitization for a pair of reactions.
    Flags:
        A: two product atoms are mapped to the same precursor (i.e. two product atoms have the same mapping number)
        B: canonicalization failed because atom-mapping number not in reactants
        C: other canonicalization issue
        S: sanitization failed
    """
    base_flags = {"A": False, "B": False, "C": False, "S": False}

    def preprocess_one(rxn, flags):
        try: # canonicalization
            can_rxn = canonicalize_rxn_smiles(rxn)
        except Exception as e:
            msg = str(e)
            if "Two product atoms mapped to the same precursor atom" in msg:
                flags["A"] = True
            elif re.search(r"Error:\s*'(\d+)' is not in list", msg):
                flags["B"] = True
            else: # should be "Canonicalization failed for input SMILES:...."
                flags["C"] = True
            return None
        try: # sanitization
            sanitized_rxn = sanitize_reaction_smiles(can_rxn)
            if sanitized_rxn is None:
                flags["S"] = True
                return None
        except Exception:
            flags["S"] = True
            return None
        return sanitized_rxn
    
    flags_ref = base_flags.copy()
    flags_pred = base_flags.copy()

    ref_processed = preprocess_one(ref_rxn, flags_ref)
    pred_processed = preprocess_one(pred_rxn, flags_pred)
    
    return PreprocessResult(
        reference=ref_processed,
        prediction=pred_processed,
        flags_ref=flags_ref,
        flags_pred=flags_pred,
    )


# Level 1 - formatting diagnostics
def diagnostic_flags_reaction(rxn_smi: str) -> Dict[str, bool]:
    """
    Run validation checks on the canonicalized, sanitized reaction.
    Returns diagnostic flags that do not block evaluation
    """
    flags = {
        "D": False, # Unmapped atom in product
        #"E": False, # eventually future diagnostics to run
    }
    if detect_atom_no_mapnum_in_preproc_prod(rxn_smi):
        flags["D"] = True
    return flags


# Level 2 - preprocessing + diagnostics
def batch_preprocess_reaction_pairs(
    pairs: Iterable[Union[ReactionPair, tuple[str, str]]]
) -> List[PreprocessResult]:
    """
    Apply preprocessing and diagnostic flagging to a list of reaction pairs.

    Args:
        pairs: Iterable of either ReactionPair objects or (ground_truth, prediction) tuples.

    Returns:
        List of PreprocessResult objects (one per pair).
    """
    results = []

    for pair in pairs:
        if isinstance(pair, ReactionPair):
            ref_rxn = pair.ground_truth
            pred_rxn = pair.prediction
        else:
            ref_rxn, pred_rxn = pair  # tuple[str, str]

        # --- step 1: preprocessing ---
        result = preprocess_reaction_pair(ref_rxn=ref_rxn, pred_rxn=pred_rxn)

        # --- step 2: diagnostics ---
        if result.reference:
            diag_flag = diagnostic_flags_reaction(result.reference)
            result.flags_ref.update(diag_flag)
        if result.prediction:
            diag_flag = diagnostic_flags_reaction(result.prediction)
            result.flags_pred.update(diag_flag)

        results.append(result)

    return results


def convert_preproc_to_df_via_list(results: List[PreprocessResult]) -> pd.DataFrame:
    """
    Converts results of preprocessing (i.e. a list of PreprocessResult instances) into pd.DataFrame.

    Args:
        results: list of PreprocessResults instances, out from `batch_preprocess_reaction_pairs()`
    
    Returns:
        pd.DataFrame, same format as original file with preprocessed reactions and flags.
    """
    data = []
    for i, r in enumerate(results):
        ref_flags = [k for k,v in r.flags_ref.items() if v]
        pred_flags = [k for k,v in r.flags_pred.items() if v]

        row = {
            "pair_index": i,
            "preproc_ref": r.reference, 
            "preproc_pred": r.prediction,
            "flags_ref": ref_flags or None,
            "flags_pred": pred_flags or None,
        }
        data.append(row)
    return pd.DataFrame(data)


def convert_preproc_to_df(results: List[PreprocessResult]) -> pd.DataFrame:
    """
    Convert a list of PreprocessResult instances into a flat, analysis-ready DataFrame.

    Each flag is expanded into its own Boolean column (ref_A, ref_B, ... / pred_A, pred_B, ...),
    allowing for direct filtering and aggregation.

    Args:
        results: List of PreprocessResult objects (output of `batch_preprocess_reaction_pairs()`)

    Returns:
        pd.DataFrame with columns:
            - pair_index: integer index of the pair
            - preproc_ref: preprocessed reference reaction
            - preproc_pred: preprocessed predicted reaction
            - ref_<FLAG>: boolean column for each flag in reference
            - pred_<FLAG>: boolean column for each flag in prediction
    """
    rows = []
    all_flags = None
    for r in results:
        if r.reference is not None or r.prediction is not None:
            all_flags = sorted({*r.flags_ref.keys(), *r.flags_pred.keys()})
            break
    if all_flags is None:
        all_flags = []

    for i, r in enumerate(results):
        row = {
            "pair_index": i,
            "preproc_ref": r.reference,
            "preproc_pred": r.prediction,
        }

        # one column per flag: True/False for ref and pred sides
        for flag in all_flags:
            row[f"ref_{flag}"] = bool(r.flags_ref.get(flag, False))
            row[f"pred_{flag}"] = bool(r.flags_pred.get(flag, False))

        rows.append(row)

    return pd.DataFrame(rows)



def preprocess_dataset(df: pd.DataFrame, path_to_save: Optional[str] = None):
    """
    Simple placeholder wrapper of the whole preprocessing approach for pd.DataFrame inputs.

    Args:
        df: pd.DataFrame that contains colunms `ground_truth_rxn`, `predicted_rxn` 
        and `row_idx` to preprocess.
        path_to_save (optional): path to file where the preprocessed dataframe as CSV should be saved

    Returns:
        pd.DataFrame: DataFrame with same format as original, with added columns describing what happened.
        Data is ready to be sorted and evaluated.
    """
    # check that columns "ground_truth_rxn", "predicted_rxn", "row_idx" exist
    if not set(["ground_truth_rxn", "predicted_rxn"]).issubset(df.columns):
        raise ValueError(f"df does not contain 'ground_truth_rxn' and 'predicted_rxn' columns, only: {df.columns}")
    
    pairs = [
        ReactionPair(row.ground_truth_rxn, row.predicted_rxn, id=row.get("row_idx"))
        for _, row in df.iterrows()
    ]
    results = batch_preprocess_reaction_pairs(pairs)
    df_preproc = convert_preproc_to_df(results)
    merged = df.merge(df_preproc, how="left", left_index=True, right_on="pair_index").drop(columns=["pair_index"])
    if path_to_save:
        merged.to_csv(path_to_save, index=False)
    return merged



import argparse
import csv
from typing import List

from .data_models import ReactionPair
from .input_io import load_pairs_from_csv, load_pairs_from_json
from .pair_evaluation import evaluate_pairs_in_parallel

"""
# Example usages
## CSV
python -m atommap_eval.cli reactions.csv -f csv -p 4 -o results.csv
## JSON
python -m atommap_eval.cli reactions.json -f json

"""


def parse_args():
    parser = argparse.ArgumentParser(
        description="Evaluate equivalence of atom-mapped reaction SMILES."
    )
    parser.add_argument("input", help="Path to input file (CSV or JSON)")
    parser.add_argument(
        "-f",
        "--format",
        choices=["csv", "json"],
        required=True,
        help="Input file format",
    )
    parser.add_argument("-o", "--output", help="Path to save output results as CSV")
    parser.add_argument(
        "-p",
        "--processes",
        type=int,
        default=None,
        help="Number of processes (default: all CPUs)",
    )
    parser.add_argument(
        "--gt-col", default="ground_truth", help="CSV column for ground truth SMILES"
    )
    parser.add_argument(
        "--pred-col", default="prediction", help="CSV column for predicted SMILES"
    )
    parser.add_argument(
        "--id-col", default="id", help="CSV column for reaction ID (optional)"
    )
    return parser.parse_args()


def load_pairs(
    input_path: str, fmt: str, gt_col: str, pred_col: str, id_col: str
) -> List[ReactionPair]:
    if fmt == "csv":
        return load_pairs_from_csv(
            input_path, gt_col=gt_col, pred_col=pred_col, id_col=id_col
        )
    elif fmt == "json":
        return load_pairs_from_json(input_path)
    else:
        raise ValueError("Unsupported format")


def write_results(pairs: List[ReactionPair], results: List[bool], output_path: str):
    with open(output_path, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["id", "ground_truth", "prediction", "equivalent"])
        for pair, result in zip(pairs, results):
            writer.writerow([pair.id or "", pair.ground_truth, pair.prediction, result])


def main():
    args = parse_args()

    print(f"Loading input pairs from: {args.input}")
    pairs = load_pairs(args.input, args.format, args.gt_col, args.pred_col, args.id_col)

    print(
        f"Running equivalence evaluation on {len(pairs)} pairs using "
        f"{args.processes or 'all available'} processes..."
    )
    results = evaluate_pairs_in_parallel(pairs, num_workers=args.processes)

    num_equiv = sum(results)
    print(f"Finished! {num_equiv} / {len(pairs)} reactions are equivalent.")

    if args.output:
        print(f"Saving results to {args.output}")
        write_results(pairs, results, args.output)


if __name__ == "__main__":
    main()



from collections import defaultdict
from typing import List

import networkx as nx
from networkx.algorithms.isomorphism import (
    categorical_edge_match,
    categorical_node_match,
)
from rdkit import Chem
from rdkit.Chem import AllChem

from .preprocess import canonicalize_rxn_smiles


def build_atom_graph(mol: Chem.Mol) -> nx.Graph:
    """
    Convert an RDKit molecule to a NetworkX graph with key atom and bond features.
    """
    G = nx.Graph()
    for atom in mol.GetAtoms():
        G.add_node(
            atom.GetIdx(),
            atomic_num=atom.GetAtomicNum(),
            atom_map=atom.GetAtomMapNum(),
            formal_charge=atom.GetFormalCharge(),
            degree=atom.GetDegree(),
        )
    for bond in mol.GetBonds():
        G.add_edge(
            bond.GetBeginAtomIdx(),
            bond.GetEndAtomIdx(),
            bond_type=str(bond.GetBondType()),
        )
    return G


def get_mapped_graphs(mols: List[Chem.Mol]) -> List[nx.Graph]:
    """
    Keep only molecule graphs with at least one atom-mapped atom.
    """
    graphs = []
    for mol in mols:
        g = build_atom_graph(mol)
        if any(g.nodes[n]["atom_map"] > 0 for n in g.nodes):
            graphs.append(g)
    return graphs


def union_labeled_graphs(
    reactants: List[nx.Graph], products: List[nx.Graph]
) -> nx.Graph:
    """
    Create a union graph with uniquely labeled nodes from reactants and products.
    """
    unionG = nx.Graph()

    def relabel_and_add(graphs, prefix):
        for i, g in enumerate(graphs):
            mapping = {n: f"{prefix}{i}-{n}" for n in g.nodes}
            renamed = nx.relabel_nodes(g, mapping)
            unionG.update(renamed)

    relabel_and_add(reactants, "R-")
    relabel_and_add(products, "P-")

    return unionG


def add_mapping_edges(unionG: nx.Graph) -> nx.Graph:
    """
    Add special edges between reactant and product atoms with the same atom_map.
    """
    atom_map_dict = nx.get_node_attributes(unionG, "atom_map")
    amap_to_nodes = defaultdict(list)
    for node, amap in atom_map_dict.items():
        if amap > 0:
            amap_to_nodes[amap].append(node)

    for amap, nodes in amap_to_nodes.items():
        rs = [n for n in nodes if n.startswith("R-")]
        ps = [n for n in nodes if n.startswith("P-")]
        for r in rs:
            for p in ps:
                unionG.add_edge(r, p, label="mapping")

    return unionG


def build_reaction_graph(rxn_smi: str, canonicalize: bool = True) -> nx.Graph:
    """
    Build a reaction graph from a reaction SMILES, optionally canonicalizing it first.

    Args:
        rxn_smi: Atom-mapped reaction SMILES
        canonicalize: Whether to canonicalize the input before processing,
        canonicalizing both SMILES and atom-map order

    Returns:
        NetworkX graph of the reaction
    """

    if not isinstance(rxn_smi, str):
        raise TypeError(f"Expected string, got {type(rxn_smi)}")

    if canonicalize:
        rxn_smi = canonicalize_rxn_smiles(rxn_smi)
    rxn = AllChem.ReactionFromSmarts(rxn_smi, useSmiles=True)

    reactant_graphs = get_mapped_graphs(rxn.GetReactants())
    product_graphs = get_mapped_graphs(rxn.GetProducts())

    union = union_labeled_graphs(reactant_graphs, product_graphs)
    return add_mapping_edges(union)


def define_node_edge_matches():
    """
    Define node and edge matchers used for graph isomorphism testing.
    """
    node_match = categorical_node_match(
        ["atomic_num", "formal_charge", "degree"], [None, None, None]
    )
    edge_match = categorical_edge_match(["bond_type", "label"], [None, None])
    return node_match, edge_match



# Functions taken from RXNMapper original in rxnmapper/smiles_utils.py available here: https://github.com/rxn4chemistry/rxnmapper
import logging
from typing import List, Tuple

from rdkit import Chem
from rxn.chemutils.conversion import smiles_to_mol
from rxn.chemutils.exceptions import InvalidSmiles
from rxn.chemutils.reaction_equation import (
    ReactionEquation,
)
from rxn.chemutils.reaction_smiles import parse_any_reaction_smiles

LOGGER = logging.getLogger(__name__)


def create_rxn_smiles_from_rxn_equation(rxn_equation) -> str:
    """
    Converts a parsed reaction equation object into a SMILES string.

    Args:
        rxn_equation: Parsed reaction equation with reactants/products.

    Returns:
        str: Reaction SMILES string.
    """
    reactants = ".".join(rxn_equation.reactants)
    products = ".".join(rxn_equation.products)
    return f"{reactants}>>{products}"


def canonicalize_and_atom_map(smi: str, return_equivalent_atoms=False):
    """
    Remove atom mapping, canonicalize and return mapping numbers
    in order of canonicalization.

    Args:
        smi: reaction SMILES str
        return_equivalent_atoms

    Returns:

    """
    mol = smiles_to_mol(smi, sanitize=False)
    for atom in mol.GetAtoms():  # type: ignore[call-arg]
        if atom.HasProp("molAtomMapNumber"):
            atom_map = atom.GetAtomMapNum()
            atom.SetProp("atom_map", str(atom_map))
            atom.ClearProp("molAtomMapNumber")
        else:
            atom.SetProp("atom_map", str(0))
    can_smi = Chem.MolToSmiles(mol)
    order = list(
        mol.GetPropsAsDict(includePrivate=True, includeComputed=True)[
            "_smilesAtomOutputOrder"
        ]
    )

    atom_maps_canonical = [mol.GetAtoms()[idx].GetProp("atom_map") for idx in order]  # type: ignore[call-arg]

    if not return_equivalent_atoms:
        return (can_smi, atom_maps_canonical)

    raise NotImplementedError


def process_reaction_with_product_maps_atoms(rxn, skip_if_not_in_precursors=False):
    """
    Remove atom-mapping, move reagents to reactants and canonicalize reaction.
    If fragment group information is given, keep the groups together using
    the character defined with fragment_bond.

    Args:
        rxn: Reaction SMILES
        skip_if_not_in_precursors: accept unmapped atoms in the product (default: False)

    Returns: joined_precursors>>joined_products reaction SMILES
    """
    reactants, reagents, products = rxn.split(">")
    try:
        precursors = [canonicalize_and_atom_map(r) for r in reactants.split(".")]
        if len(reagents) > 0:
            precursors += [canonicalize_and_atom_map(r) for r in reagents.split(".")]
        products = [canonicalize_and_atom_map(p) for p in products.split(".")]
    except InvalidSmiles:
        return ""
    sorted_precursors = sorted(precursors, key=lambda x: x[0])
    sorted_products = sorted(products, key=lambda x: x[0])
    joined_precursors = ".".join([p[0] for p in sorted_precursors])
    joined_products = ".".join([p[0] for p in sorted_products])
    precursors_atom_maps = [
        i for p in sorted_precursors for i in p[1]
    ]  # could contain duplicate entries
    product_atom_maps = [
        i for p in sorted_products for i in p[1]
    ]  # could contain duplicate entries

    joined_rxn = f"{joined_precursors}>>{joined_products}"

    products_maps = []
    warnings = []

    for p_map in product_atom_maps:
        if skip_if_not_in_precursors and p_map not in precursors_atom_maps:
            products_maps.append(-1)
        elif int(p_map) == 0:
            products_maps.append(-1)
        else:
            corresponding_precursors_atom = precursors_atom_maps.index(p_map)
            if (
                corresponding_precursors_atom in products_maps
            ):  # handle equivalent atoms
                found_alternative = False
                for atom_idx, precursor_map in enumerate(precursors_atom_maps):
                    if (precursor_map == p_map) and atom_idx not in products_maps:
                        products_maps.append(atom_idx)
                        found_alternative = True
                        break
                if not found_alternative:
                    msg = f"Two product atoms mapped to the same precursor atom: {rxn}"
                    warnings.append(msg)
                    products_maps.append(corresponding_precursors_atom)
            else:
                products_maps.append(corresponding_precursors_atom)
    for w in list(set(warnings)):
        LOGGER.warning(w)
        raise ValueError(w)
    return joined_rxn, products_maps


def generate_atom_mapped_reaction_atoms(
    rxn: str, product_atom_maps, expected_atom_maps=None, canonical: bool = False
) -> Tuple[ReactionEquation, List[int]]:
    """
    Generate atom-mapped reaction from unmapped reaction and
    product-2-reactant atoms mapping vector.
    Args:
        rxn: unmapped reaction, in the format that the transformer model relies on.
        product_atom_maps: product to reactant atom maps.
        expected_atom_maps: if given, return the differences.
        canonical: whether to canonicalize the resulting SMILES.

    Returns: Atom-mapped reaction

    """

    reactants, agents, products = parse_any_reaction_smiles(rxn)
    precursors_mols = [smiles_to_mol(pr, sanitize=False) for pr in reactants + agents]
    products_mols = [smiles_to_mol(prod, sanitize=False) for prod in products]

    precursors_atom_maps = []

    differing_maps = []

    product_mapping_dict = {}

    i = -1
    atom_mapped_precursors_list = []
    for precursor_mol in precursors_mols:
        for atom in precursor_mol.GetAtoms():  # type: ignore[call-arg]
            i += 1
            if i in product_atom_maps:
                # atom maps start at an index of 1
                corresponding_product_atom_map = product_atom_maps.index(i) + 1
                precursors_atom_maps.append(corresponding_product_atom_map)
                atom.SetProp("molAtomMapNumber", str(corresponding_product_atom_map))

                indices = [idx for idx, x in enumerate(product_atom_maps) if x == i]

                if len(indices) > 1:
                    for idx in indices[1:]:
                        product_mapping_dict[idx] = corresponding_product_atom_map

                if expected_atom_maps is not None:
                    if (
                        i not in expected_atom_maps
                        or corresponding_product_atom_map
                        != expected_atom_maps.index(i) + 1
                    ):
                        differing_maps.append(corresponding_product_atom_map)
        atom_mapped_precursors_list.append(
            Chem.MolToSmiles(precursor_mol, canonical=canonical)
        )

    i = -1
    atom_mapped_products_list = []
    for products_mol in products_mols:
        for atom in products_mol.GetAtoms():  # type: ignore[call-arg]
            i += 1
            atom_map = product_mapping_dict.get(i, i + 1)
            atom.SetProp("molAtomMapNumber", str(atom_map))
        atom_mapped_products_list.append(
            Chem.MolToSmiles(products_mol, canonical=canonical)
        )

    atom_mapped_rxn = ReactionEquation(
        atom_mapped_precursors_list, [], atom_mapped_products_list
    )

    return atom_mapped_rxn, differing_maps
